# Tai Sakuma <tai.sakuma@gmail.com>
import time

##__________________________________________________________________||
class Presentation(object):
    """A base class of the progress presentation.

    A subclass of this class should implement ``_present()``.
    """

    def __init__(self):

        self._new_taskids = [ ]
        self._active_taskids = [ ] # in order of arrival
        self._finishing_taskids = [ ]
        self._complete_taskids = [ ] # in order of completion
        self._report_dict = { }

        self.interval = 1.0 # [second]
        self._read_time()

    def active(self):
        if self._active_taskids:
            return True
        return False

    def present(self, report):

        if not self._register_report(report):
            return

        if not self._need_to_present():
            return

        self._present()

        self._update_registry()

        self._read_time()

    def _register_report(self, report):

        if report['taskid'] in self._complete_taskids:
            return False

        self._report_dict[report['taskid']] = report

        if report['taskid'] in self._finishing_taskids:
            return True

        if report['last']:
            try:
                self._active_taskids.remove(report['taskid'])
            except ValueError:
                pass

            try:
                self._new_taskids.remove(report['taskid'])
            except ValueError:
                pass

            self._finishing_taskids.append(report['taskid'])

            return True

        if report['taskid'] in self._active_taskids:
            return True

        if report['taskid'] in self._new_taskids:
            return True

        self._new_taskids.append(report['taskid'])
        return True

    def _update_registry(self):
        self._active_taskids.extend(self._new_taskids)
        del self._new_taskids[:]

        self._complete_taskids.extend(self._finishing_taskids)
        del self._finishing_taskids[:]

    def _need_to_present(self):

        if self._new_taskids:
            return True

        if self._finishing_taskids:
            return True

        if self._time() - self.last_time > self.interval:
            return True

        return False

    def _time(self):
        return time.time()

    def _read_time(self):
        self.last_time = self._time()

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import time, random
import collections

import pytest

from atpbar import atpbar
from atpbar import flush

##__________________________________________________________________||
@pytest.mark.parametrize('method', ['break', 'exception'])
@pytest.mark.parametrize('niterations', [10, 1, 0])
@pytest.mark.parametrize('ndones', [10, 4, 1, 0])
def test_one_loop_break_exception(mock_create_presentation, wrap_end_pickup, niterations, ndones, method):
    ndones = min(ndones, niterations)

    def task_break(ndones, niterations):
        for i in atpbar(range(niterations)):
            if i == ndones:
                break

    def task_exception(ndones, niterations):
        for i in atpbar(range(niterations)):
            if i == ndones:
                raise Exception()

    #
    if method == 'break':
        task_break(ndones, niterations)
    else:
        if ndones < niterations:
            with pytest.raises(Exception):
                task_exception(ndones, niterations)
        else:
            task_exception(ndones, niterations)

    ## print()
    ## print(mock_create_presentation)

    #
    assert 1 == wrap_end_pickup.call_count

    nreports_expected = ndones + 1 + bool(ndones<niterations)
    presentations = mock_create_presentation.presentations

    assert 2 == len(presentations) # created when atpbar started and ended

    #
    progressbar0 = presentations[0]
    assert nreports_expected == len(progressbar0.reports)
    assert 1 == len(progressbar0.taskids)
    assert 1 == progressbar0.nfirsts
    assert 1 == progressbar0.nlasts
    done_total_list_expected = [(ndones, niterations)]
    done_total_list_actual = [(d['done'], d['total']) for d in progressbar0._report_dict.values()]
    assert collections.Counter(done_total_list_expected) == collections.Counter(done_total_list_actual)

    #
    progressbar1 = presentations[1]
    assert 0 == len(progressbar1.reports)

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import sys
import atexit
import threading
import multiprocessing
import contextlib

from .reporter import ProgressReporter
from .pickup import ProgressReportPickup
from .presentation.create import create_presentation
from . import detach

##__________________________________________________________________||
_lock = threading.Lock()
_queue = None
_reporter = None
_pickup = None
_pickup_owned = False
_do_not_start_pickup = False

##__________________________________________________________________||
def find_reporter():
    """returns the progress reporter

    This function is to be called in the main process of a
    multiprocessing program. The reporter should be registered in
    sub-processes with the function `register_reporter()`

    Returns
    -------
    object
        The progress reporter

    """

    global _lock
    global _reporter

    _lock.acquire()
    _start_pickup_if_necessary()
    _lock.release()

    return _reporter

##__________________________________________________________________||
def register_reporter(reporter):
    """registers a reporter

    This function is to be called in sub-processes of a
    multiprocessing program.

    Parameters
    ----------
    reporter : object
        The reporter obtained in the main process by the function
        `find_reporter()`


    Returns
    -------
    None

    """

    global _reporter
    global _do_not_start_pickup
    _reporter = reporter
    _do_not_start_pickup = True

##__________________________________________________________________||
def flush():
    """flushes progress bars

    This function flushes all active progress bars. It returns when
    the progress bars finish updating.

    Returns
    -------
    None

    """
    global _lock
    _lock.acquire()
    _end_pickup()
    _start_pickup_if_necessary()
    _lock.release()

##__________________________________________________________________||
def disable():
    """disables progress bars

    This function needs to be called in the main process before
    `atpbar()` or `find_reporter()` is used.

    Returns
    -------
    None

    """
    global _do_not_start_pickup
    _do_not_start_pickup = True

##__________________________________________________________________||
def end_pickup():
    """ends the pickup

    Returns
    -------
    None

    """
    global _lock
    _lock.acquire()
    _end_pickup()
    _lock.release()

atexit.register(end_pickup)

##__________________________________________________________________||
@contextlib.contextmanager
def fetch_reporter():
    global _lock
    global _reporter
    global _do_not_start_pickup
    global _pickup_owned

    _lock.acquire()
    _start_pickup_if_necessary()
    _lock.release()

    own_pickup = False
    if not _do_not_start_pickup:
        if in_main_thread():
            if not _pickup_owned:
                own_pickup = True
                _pickup_owned = True


    try:
        yield _reporter
    finally:
        _lock.acquire()
        if detach.to_detach_pickup:
            if own_pickup:
                own_pickup = False
                _pickup_owned = False
        if own_pickup:
            _end_pickup()
            _start_pickup_if_necessary()
        _lock.release()

def in_main_thread():
    try:
        return threading.current_thread() == threading.main_thread()
    except:
        # python 2
        return isinstance(threading.current_thread(), threading._MainThread)

##__________________________________________________________________||
def _start_pickup_if_necessary():
    global _reporter
    global _queue
    global _pickup
    global _pickup_owned
    global _do_not_start_pickup

    if _do_not_start_pickup:
        return

    if _reporter is None:
        if _queue is None:
            _queue = multiprocessing.Queue()
        _reporter = ProgressReporter(queue=_queue)

    if _pickup is not None:
        return

    presentation = create_presentation()
    _pickup = ProgressReportPickup(_queue, presentation)
    _pickup.daemon = True # this makes the functions
                          # registered at atexit called even
                          # if the pickup is still running

    _pickup.start()
    _pickup_owned = False

    return

##__________________________________________________________________||
def _end_pickup():
    global _queue
    global _pickup
    if _pickup:
        _queue.put(None)
        _pickup.join()
        _pickup = None
        detach.to_detach_pickup = False

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import time, random
import itertools
import multiprocessing

import pytest

from atpbar import atpbar
from atpbar import register_reporter, find_reporter, flush

##__________________________________________________________________||
def run_with_multiprocessing(nprocesses, ntasks, niterations, time_starting_task):

    def task(n, name, time_starting):
        time.sleep(time_starting)
        for i in atpbar(range(n), name=name):  # `atpbar` is used here
            time.sleep(0.0001)

    def worker(reporter, task, queue):
        register_reporter(reporter)
        while True:
            args = queue.get()
            if args is None:
                queue.task_done()
                break
            task(*args)
            queue.task_done()

    reporter = find_reporter()
    queue = multiprocessing.JoinableQueue()

    for i in range(nprocesses):
        p = multiprocessing.Process(target=worker, args=(reporter, task, queue))
        p.start()

    for i in atpbar(range(ntasks)): # `atpbar` is used here
        name = 'task {}'.format(i)
        n = niterations[i]
        queue.put((n, name, time_starting_task))
        time.sleep(0.01)

    for i in range(nprocesses):
        queue.put(None)
        queue.join()

    flush()

@pytest.mark.xfail()
@pytest.mark.parametrize('time_starting_task', [0, 0.01, 0.2])
@pytest.mark.parametrize('niterations', [[5, 4, 3], [5, 0, 1], [0], [1]])
@pytest.mark.parametrize('ntasks', [3, 1, 0])
@pytest.mark.parametrize('nprocesses', [6, 2, 1])
def test_multiprocessing_from_loop(
        mock_create_presentation, wrap_end_pickup,
        nprocesses, ntasks, niterations, time_starting_task):

    # make niterations as long as ntasks. repeat if necessary
    niterations = list(itertools.chain(*itertools.repeat(niterations, ntasks//len(niterations)+1)))[:ntasks]

    run_with_multiprocessing(nprocesses, ntasks, niterations, time_starting_task)

    ## print()
    ## print(mock_create_presentation)

    nreports_expected_from_main = ntasks + 1
    nreports_expected_from_tasks = sum(niterations) + ntasks
    nreports_expected = nreports_expected_from_main + nreports_expected_from_tasks

    presentations = mock_create_presentation.presentations

    if nreports_expected_from_tasks == 0:
        assert 3 == len(presentations) # in find_reporter(), at the
                                       # end of `atpbar` in the main
                                       # process, and in flush().

        progressbar0 = presentations[0]
        assert nreports_expected == len(progressbar0.reports)
        # one report from `atpbar` in the main thread

        assert 1 == progressbar0.nfirsts
        assert 1 == progressbar0.nlasts
        assert 1 == len(progressbar0.taskids)

    else:
        if 2 == len(presentations):

            progressbar1 = presentations[1]
            assert 0 == len(progressbar1.reports)

            progressbar0 = presentations[0]
            assert ntasks + 1 == len(progressbar0.taskids)
            assert ntasks + 1 == progressbar0.nfirsts
            assert ntasks + 1 == progressbar0.nlasts
            assert nreports_expected == len(progressbar0.reports)

        else:
            assert 3 == len(presentations)

            progressbar2 = presentations[2]
            assert 0 == len(progressbar2.reports)

            progressbar0 = presentations[0]
            progressbar1 = presentations[1]

            assert ntasks + 1 == len(progressbar0.taskids) + len(progressbar1.taskids)
            assert ntasks + 1 == progressbar0.nfirsts + progressbar1.nfirsts
            assert ntasks + 1 == progressbar0.nlasts + progressbar1.nlasts
            assert nreports_expected == len(progressbar0.reports) + len(progressbar1.reports)

    # At this point the pickup shouldn't be owned. Therefore, a new
    # `atpbar` in the main thread should own it.
    npresentations = len(presentations)
    for i in atpbar(range(4)):
        pass
    assert npresentations + 1 == len(presentations)
    progressbar = presentations[-2]
    assert 1 == len(progressbar.taskids)
    assert 1 == progressbar.nfirsts
    assert 1 == progressbar.nlasts
    assert 4 + 1 == len(progressbar.reports)

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import pytest

has_jupyter_notebook = False
try:
    import ipywidgets as widgets
    from IPython.display import display
    has_jupyter_notebook = True
except ImportError:
    pass

from atpbar.presentation.bartty import ProgressBar
from atpbar.presentation.txtprint import ProgressPrint

if has_jupyter_notebook:
    from atpbar.presentation.barjupyter import ProgressBarJupyter

##__________________________________________________________________||
classes = [ProgressBar, ProgressPrint]

if has_jupyter_notebook:
    classes.append(ProgressBarJupyter)

classe_ids = [c.__name__ for c in classes]

@pytest.mark.parametrize('Class', classes, ids=classe_ids)
def test_presentation(Class):
    obj = Class()
    repr(obj)
    obj.active()
    obj.present(dict(name='task1', done=0, total=10, taskid=1, first=True, last=False))
    obj.present(dict(name='task1', done=2, total=10, taskid=1, first=False, last=False))
    obj.active()
    obj.present(dict(name='task1', done=0, total=5, taskid=2, first=True, last=False))
    obj.present(dict(name='task1', done=3, total=5, taskid=2, first=False, last=False))
    obj.active()
    obj.present(dict(name='task1', done=10, total=10, taskid=1, first=False, last=True))
    obj.active()
    obj.present(dict(name='task1', done=5, total=5, taskid=2, first=False, last=True))
    obj.active()
    obj.active()
    obj.present(dict(name='task1', done=10, total=10, taskid=1, first=False, last=True))
    obj.active()

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import sys
import threading

import pytest

try:
    import unittest.mock as mock
except ImportError:
    import mock

from atpbar.presentation.base import Presentation

from atpbar.funcs import end_pickup, _end_pickup

##__________________________________________________________________||
class MockProgressBar(Presentation):
    def __init__(self):
        super(MockProgressBar, self).__init__()
        self.reports = [ ]
        self.taskids = set()
        self.nfirsts = 0
        self.nlasts = 0

    def __str__(self):
        lines = [ ]
        l = '{}: # reports: {}, # taskids: {}, # firsts: {}, # lasts: {}'.format(
            self.__class__.__name__,
            len(self.reports), len(self.taskids), self.nfirsts, self.nlasts
        )
        lines.append(l)
        lines.extend(['  {}'.format(r) for r in self.reports])
        return '\n'.join(lines)

    def present(self, report):
        super(MockProgressBar, self).present(report)
        self.reports.append(report)
        self.taskids.add(report['taskid'])
        self.nfirsts += report['first']
        self.nlasts += report['last']

    def _present(self):
        pass

class MockCreatePresentation(object):
    """A functor to mock `create_presentation()`.

    It keeps returned values so they can be examined later.
    """
    def __init__(self):
        self.presentations = [ ]

    def __str__(self):
        lines = [ ]
        lines.append('{}:'.format(self.__class__.__name__))
        lines.extend(['  {}'.format('\n  '.join(str(p).split('\n'))) for p in self.presentations])
        return '\n'.join(lines)

    def __call__(self):
        ret = MockProgressBar()
        self.presentations.append(ret)
        return ret

@pytest.fixture(autouse=True)
def mock_create_presentation(monkeypatch):
    ret = MockCreatePresentation()
    module = sys.modules['atpbar.funcs']
    monkeypatch.setattr(module, 'create_presentation', ret)
    return ret

##__________________________________________________________________||
@pytest.fixture(autouse=True)
def global_variables(monkeypatch):
    module = sys.modules['atpbar.funcs']
    monkeypatch.setattr(module, '_lock', threading.Lock())
    monkeypatch.setattr(module, '_queue', None)
    monkeypatch.setattr(module, '_reporter', None)
    monkeypatch.setattr(module, '_pickup', None)
    monkeypatch.setattr(module, '_pickup_owned', False)
    monkeypatch.setattr(module, '_do_not_start_pickup', False)

    module = sys.modules['atpbar.detach']
    monkeypatch.setattr(module, 'to_detach_pickup', False)

    yield

    end_pickup()

##__________________________________________________________________||
@pytest.fixture(autouse=True)
def wrap_end_pickup(monkeypatch):
    ret = mock.Mock(wraps=_end_pickup)
    module = sys.modules['atpbar.funcs']
    monkeypatch.setattr(module, '_end_pickup', ret)
    yield ret

##__________________________________________________________________||
@pytest.fixture(autouse=True)
def reporter_interval(monkeypatch):
    module = sys.modules['atpbar.reporter']
    monkeypatch.setattr(module, 'DEFAULT_INTERVAL', 0)
    yield

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>

##__________________________________________________________________||
class ProgressReportComplementer(object):
    """Complement progress reports

    Complement a progress report with the previous report for the same
    task.

    Parameters
    ----------
    report : dict
        A progress report, a dict with the following entries. The
        `taskid` must be always given. The first report for a task
        must include `done`, `total`, 'name', 'pid', and
        'in_main_thread`. The `first` and `last` will be automatically
        determined if not given.

        taskid : immutable
            The unique task ID.
        done : int, optional
            The number of the iterations done so far
        total : int
            The total iterations to be done
        name : str
            A name of the task. It will be use as the label on the
            progress bars.
        pid : optional
            The process ID, e.g., the value returned by `os.getpid()`
        in_main_thread : bool
            `True` if the task is running in the main thread of the
            process. `False` otherwise
        first : bool
            `True` if the first report for the task. If not given,
            automatically determined from `done`; `True` if `done` is
            0, `False` otherwise
        last : bool
            `True` if the last report for the task. If not given,
            automatically determined from `done` and `total`; `True`
            if `done` equals `total`, `False` otherwise

    """
    def __init__(self):
        self.previous_reports = { }
        self.volatile_fileds = ('first', 'last')

    def __call__(self, report):
        taskid = report['taskid']
        if taskid in self.previous_reports:
            self._complement(taskid, report)
        self._first(report)
        self._last(report)
        self._store(taskid, report.copy())

    def _complement(self, taskid, report):
        report_copy = report.copy()
        report.clear()
        report.update(self.previous_reports[taskid])
        report.update(report_copy)

    def _first(self, report):
        if 'first' in report:
            return
        report['first'] = (report['done'] == 0)

    def _last(self, report):
        if 'last' in report:
            return
        report['last'] = (report['done'] >= report['total'])

    def _store(self, taskid, report):
        for k in self.volatile_fileds:
            report.pop(k, None)
        self.previous_reports[taskid] = report

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import sys, time

from .base import Presentation

##__________________________________________________________________||
class ProgressPrint(Presentation):
    def __init__(self):
        super(ProgressPrint, self).__init__()
        self.interval = 60.0 # [second]
        self.last_time = { } # key: taskid

    def __repr__(self):
        return '{}()'.format(
            self.__class__.__name__
        )

    def present(self, report):

        if not self._register_report(report):
            return

        if not self._need_to_present(report):
            return

        self._present(report)

        self.last_time[report['taskid']] = self._time()

    def _present(self, report):
        time_ = time.strftime('%m/%d %H:%M', time.localtime(time.time()))
        percent = float(report['done'])/report['total'] if report['total'] > 0 else 1
        percent = round(percent * 100, 2)
        line = "{time} : {done:8d} / {total:8d} ({percent:6.2f}%): {name} ".format(
            time=time_,
            done=report['done'], total=report['total'],
            percent=percent, name=report['name']
        )
        line = '{}\n'.format(line)
        sys.stdout.write(line)
        sys.stdout.flush()

    def _need_to_present(self, report):

        if report['first']:
            return True

        if report['last']:
            return True

        if report['taskid'] not in self.last_time:
            return True

        if self._time() - self.last_time[report['taskid']] > self.interval:
            return True

        return False

    def _time(self):
        return time.time()

##__________________________________________________________________||
